#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <omp.h>
#include <new>
#include "Timer.h"

int main(int n_arg_num, const char **p_arg_list)
{
	bool b_verbose = true;
	size_t n_vector_length = 1 << 26;
	// input parameters

	int *p_vector_a, *p_vector_b, *p_vector_c, *p_para_result;
	if(!(p_vector_a = new(std::nothrow) int[n_vector_length]) ||
	   !(p_vector_b = new(std::nothrow) int[n_vector_length]) ||
	   !(p_vector_c = new(std::nothrow) int[n_vector_length]) ||
	   !(p_para_result = new(std::nothrow) int[n_vector_length])) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// allocate buffers

	for(size_t i = 0; i < n_vector_length; ++ i) {
		p_vector_a[i] = rand();
		p_vector_b[i] = rand();
	}
	// generate random input data

	CTimer timer;
	timer.ResetTimer();

	double f_serial_start = timer.f_Time();

	{
		for(size_t i = 0; i < n_vector_length; ++ i)
			p_vector_c[i] = p_vector_a[i] + p_vector_b[i];
		// run the calculation on CPU
	}
	// CPU version of the algorithm

	double f_serial_time = timer.f_Time() - f_serial_start;

	double f_parallel_start = timer.f_Time();

	{
		//omp_set_num_threads(4);
		// this is how to tell OpenMP how many threads to use

#if 1
		#pragma omp parallel for
		for(int i = 0; i < n_vector_length; ++ i)
			p_para_result[i] = p_vector_a[i] + p_vector_b[i];
#else
		#pragma omp parallel
		{
			int nt = omp_get_num_threads();
			int tid = omp_get_thread_num();
			size_t b = n_vector_length / nt * tid;
			size_t e = (tid + 1 == nt)? n_vector_length : n_vector_length / nt * (tid + 1);
			for(size_t i = b; i < e; ++ i)
				p_para_result[i] = p_vector_a[i] + p_vector_b[i];
		}
#endif
		// run the calculation on CPU
	}
	// CPU version of the algorithm

	double f_parallel_time = timer.f_Time() - f_parallel_start;

	if(!memcmp(p_para_result, p_vector_c, n_vector_length * sizeof(int)))
		printf("ok, parallel results match serial\n");
	else
		fprintf(stderr, "error: results calcualted on parallel do not match serial results\n");
	// check the results

	if(b_verbose)
		printf("finished ...\n");
	// verbose

	printf("serial time: %f\n", f_serial_time);
	printf("parallel time: %f\n", f_parallel_time);
	// results

	delete[] p_vector_a;
	delete[] p_vector_b;
	delete[] p_vector_c;
	delete[] p_para_result;

	return 0;
}
